@@ -25,11 +25,15 @@ from jottalib import __version__
 # importing stdlib
 import sys, os, os.path, time
 import urllib, logging, datetime, hashlib
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from StringIO import StringIO
 
 # importing external dependencies (pip these, please!)
 import requests
 import requests_cache
-requests_cache.install_cache('jfs', backend='sqlite', expire_after=100.0, fast_save=True)
+requests_cache.core.install_cache(backend='memory', expire_after=30.0, fast_save=True)
 import lxml, lxml.objectify
 import dateutil, dateutil.parser
 
@@ -40,31 +44,36 @@ logging.basicConfig(level=logging.DEBUG)
 
 class JFSError(Exception):
     @staticmethod
-    def raiseError(e, path): # parse object from lxml.objectify and 
-        if(e.code) == 400:
-            raise JFSBadRequestError('Bad request: %s (%s)' % (path, e.message))
-        elif(e.code) == 404:
+    def raiseError(e, path): # parse object from lxml.objectify and
+        if(e.code) == 404:
             raise JFSNotFoundError('%s does not exist (%s)' % (path, e.message))
+        elif(e.code) == 401:
+            raise JFSCredentialsError("Your credentials don't match for %s (%s) (probably incorrect password!)" % (path, e.message))
         elif(e.code) == 403:
             raise JFSAuthenticationError("You don't have access to %s (%s)" % (path, e.message))
         elif(e.code) == 500:
             raise JFSServerError("Internal server error: %s (%s)" % (path, e.message))
+        elif(e.code) == 400:
+            raise JFSBadRequestError('Bad request: %s (%s)' % (path, e.message))
         else:
             raise JFSError('Error accessing %s (%s)' % (path, e.message))
 
-class JFSBadRequestError(JFSError):
+class JFSBadRequestError(JFSError): # HTTP 400
     pass
 
-class JFSNotFoundError(JFSError):
+class JFSCredentialsError(JFSError): # HTTP 401
     pass
 
-class JFSAccessError(JFSError):
+class JFSNotFoundError(JFSError): # HTTP 404
     pass
 
-class JFSAuthenticationError(JFSError):
+class JFSAccessError(JFSError): #
     pass
 
-class JFSServerError(JFSError):
+class JFSAuthenticationError(JFSError): # HTTP 403
+    pass
+
+class JFSServerError(JFSError): # HTTP 500
     pass
 
 class JFSFolder(object):
@@ -131,17 +140,26 @@ class JFSFolder(object):
         self.sync()
         return r
 
-    def up(self, fileobj_or_path):
+    def rename(self, newpath):
+        "Move folder to a new name, possibly a whole new path"
+        # POST https://www.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/Ny%20mappe?mvDir=/**USERNAME**/Jotta/Sync/testFolder
+        url = '%s?mvDir=/%s%s' % (self.path, self.jfs.username, newpath)
+        r = self.jfs.post(url, extra_headers={'Content-Type':'application/octet-stream'})
+        return r
+
+    def up(self, fileobj_or_path, filename=None):
         'Upload a file to current folder and return the new JFSFile'
         if not isinstance(fileobj_or_path, file):
             fileobj_or_path = open(fileobj_or_path, 'rb')
-        r = self.jfs.up('%s/%s' % (self.path, os.path.basename(fileobj_or_path.name)), fileobj_or_path)
+        if filename is None:
+            filename = os.path.basename(fileobj_or_path.name)
+        r = self.jfs.up('%s/%s' % (self.path, filename), fileobj_or_path)
         self.sync()
         return r
 
 class JFSFile(object):
     'OO interface to a file, for convenient access. Type less, do more.'
-    ## TODO: add <revisions> iterator for all 
+    ## TODO: add <revisions> iterator for all
     """
 <file name="jottacloud.sync.pdfname" uuid="37530f11-d55b-4f31-acf4-27854813cd34" time="2013-12-15-T01:11:52Z" host="dn-029.site-000.jotta.no">
   <path xml:space="preserve">/havardgulldahl/Jotta/Sync</path>
@@ -193,8 +211,11 @@ class JFSFile(object):
 
     def write(self, data):
         'Put, possibly replace, file contents with (new) data'
-        return self.jfs.post(self.path, contents=data, 
-                             extra_headers={'Content-Type':'application/octet-stream'})
+#         return self.jfs.post(self.path, content=data,
+#                              extra_headers={'Content-Type':'application/octet-stream'})
+        if not hasattr(data, 'read'):
+            data = StringIO(data)
+        self.jfs.up(self.path, data)
 
     def share(self):
         'Enable public access at secret, share only uri, and return that uri'
@@ -212,6 +233,13 @@ class JFSFile(object):
         r = self.jfs.post(url)
         return r
 
+    def rename(self, newpath):
+        "Move file to a new name, possibly a whole new path"
+        # POST https://www.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/testFolder/testFile.txt?mv=/**USERNAME**/Jotta/Sync/testFolder/renamedTestFile.txt
+        url = '%s?mv=/%s%s' % (self.path, self.jfs.username, newpath)
+        r = self.jfs.post(url, extra_headers={'Content-Type':'application/octet-stream'})
+        return r
+
     def thumb(self, size=BIGTHUMB):
         'Get a thumbnail as string or None if the file isnt an image'
         if not self.is_image():
@@ -452,34 +480,35 @@ class JFSenableSharing(object):
     def sharedFiles(self):
         'iterate over shared files and get their public URI'
         for f in self.sharing.files.iterchildren():
-            yield (f.attrib['name'], f.attrib['uuid'], 
+            yield (f.attrib['name'], f.attrib['uuid'],
                 'http://www.jottacloud.com/p/%s/%s' % (self.jfs.username, f.publicURI.text))
 
 
 class JFS(object):
     def __init__(self, username, password, ca_bundle=True):
         from requests.auth import HTTPBasicAuth
-        self.ca_bundle = ca_bundle
-        self.username = username
-        self.auth = HTTPBasicAuth(username, password)
-        self.path = JFS_ROOT + username
         self.apiversion = '2.2' # hard coded per october 2014
-        self.headers =  {'User-Agent':'JottaFS %s (https://gitorious.org/jottafs/)' % (__version__, ),
-                         'X-JottaAPIVersion': self.apiversion,
-                         #'From': __author__
-                         } 
-        self.fs = self.get(self.path)
+        self.session = requests.Session() # create a session for connection pooling, ssl keepalives and cookie jar
+        self.username = username
+        self.session.auth = HTTPBasicAuth(username, password)
+        self.session.verify = ca_bundle
+        self.session.headers =  {'User-Agent':'JottaFS %s (https://gitorious.org/jottafs/)' % (__version__, ),
+                                 'X-JottaAPIVersion': self.apiversion,
+                                }
+        self.rootpath = JFS_ROOT + username
+        self.fs = self.get(self.rootpath)
 
     def request(self, url, usecache=True):
         if not url.startswith('http'):
             # relative url
-            url = self.path + url
+            url = self.rootpath + url
         logging.debug("getting url: %s" % url)
         if usecache:
-            r = requests.get(url, headers=self.headers, auth=self.auth, verify=self.ca_bundle)
+            r = self.session.get(url)
         else:
             with requests_cache.disabled():
-                r = requests.get(url, headers=self.headers, auth=self.auth, verify=self.ca_bundle)
+                r = self.session.get(url)
+
         if r.status_code in ( 500, ):
             raise JFSError(r.reason)
         return r
@@ -528,14 +557,17 @@ class JFS(object):
 
     def post(self, url, content='', files=None, params=None, extra_headers={}):
         'HTTP Post files[] or content (unicode string) to url'
-        if not url.startswith('http'):  
+        if not url.startswith('http'):
             # relative url
-            url = self.path + url
+            url = self.rootpath + url
+        logging.debug('yanking url from cache: %s', url)
+        cache = requests_cache.core.get_cache()
+        cache.delete_url(url)
         logging.debug('posting content (len %s) to url %s', content is not None and len(content) or '?', url)
-        headers = self.headers.copy()
+        headers = self.session.headers.copy()
         headers.update(**extra_headers)
-        r = requests.post(url, data=content, params=params, files=files, headers=headers, auth=self.auth, verify=self.ca_bundle)
-        if r.status_code in ( 500, 404, 401, 403 ):
+        r = self.session.post(url, data=content, params=params, files=files, headers=headers)
+        if r.status_code in ( 500, 404, 401, 403, 400 ):
             logging.warning('HTTP POST failed: %s', r.text)
             raise JFSError(r.reason)
         return self.getObject(r) # return a JFS* class
@@ -546,7 +578,7 @@ class JFS(object):
 
         *** WHAT DID I DO?: created file
         ***
-         
+
         POST https://up.jottacloud.com/jfs/**USERNAME**/Jotta/Sync/testFolder/testFile.txt?cphash=d41d8cd98f00b204e9800998ecf8427e HTTP/1.1
         User-Agent: Desktop_Jottacloud 3.0.22.203 Windows_8 6.2.9200 x86_64
         Authorization: Basic ******************
@@ -570,11 +602,11 @@ class JFS(object):
         fileobject.seek(0) # rewind read index for requests.post
         md5hash = hashlib.md5(content).hexdigest()
         logging.debug('posting content (len %s, hash %s) to url %s', len(content), md5hash, url)
-        now = '2014-10-05T10:23:18Z+00:00' #datetime.datetime.now().isoformat()
+        now = '2014-10-05T10:23:18Z+00:00' #datetime.datetime.now().isoformat() # TODO: don't hardcode
         headers = {'JMd5':md5hash,
                    'JCreated': now,
                    'JModified': now,
-                   'X-Jfs-DeviceName': 'Jotta', 
+                   'X-Jfs-DeviceName': 'Jotta',
                    'JSize': len(content),
                    'jx_csid': '',
                    'jx_lisence': ''
@@ -586,7 +618,6 @@ class JFS(object):
                  'file': (os.path.basename(url), fileobject, 'application/octet-stream', {'Content-Transfer-Encoding':'binary'})}
         return self.post(url, None, files=files, params=params, extra_headers=headers)
 
-
     # property overloading
     @property
     def devices(self):
@@ -625,7 +656,7 @@ if __name__=='__main__':
     http_client.HTTPConnection.debuglevel = 1
     requests_log = logging.getLogger("requests.packages.urllib3")
     requests_log.setLevel(logging.DEBUG)
-    requests_log.propagate = True    
+    requests_log.propagate = True
     from lxml.objectify import dump as xdump
     from pprint import pprint
     jfs = JFS(os.environ['JOTTACLOUD_USERNAME'], password=os.environ['JOTTACLOUD_PASSWORD'])
@@ -635,4 +666,4 @@ if __name__=='__main__':
         if j.name == 'Jotta':
             jottadev = j
     jottasync = jottadev.mountPoints['Sync']
-    r = jottasync.up('/tmp/test.pdf')    
+    r = jottasync.up('/tmp/test.pdf')
